
#include <iostream>
#include "Intellivision.h"

using namespace std;

Intellivision::Intellivision()
    : currentCartridge(NULL),
      periphCount(0),
      intellivoiceInUse(FALSE),
      ecsInUse(FALSE),
      playersSwitched(FALSE)
{
    init();
}

Intellivision::~Intellivision()
{
    //delete the audio output lines
    am.releaseLine(&psgAudioLine);
    am.releaseLine(&ecsAudioLine);
    am.releaseLine(&ivoiceAudioLine);

    delete player1Controller;
    delete player2Controller;
}

void Intellivision::setNetworkGame(BOOL b)
{
    emuDirector.setNetworkGame(b);
}

void Intellivision::setPlayersSwitched(BOOL b)
{
    if (playersSwitched == b)
        return;

    playersSwitched = b;
    HandController* hold = player1Controller;
    player1Controller = player2Controller;
    player2Controller = hold;
    psg.setIODevice0(player1Controller);
    psg.setIODevice1(player2Controller);
}

BOOL Intellivision::arePlayersSwitched()
{
    return playersSwitched;
}

void Intellivision::setPlayerOneRemote(BOOL b)
{
    emuDirector.setPlayerOneRemote(b);
}

void Intellivision::setPlayerTwoRemote(BOOL b)
{
    emuDirector.setPlayerTwoRemote(b);
}

/**
 * Initializes all of the basic hardware included in the Intellivision
 * Master Component as well as the ECS and Intellivoice peripherals.
 * This method is called only onces when the Intellivision is constructed.
 */
void Intellivision::init() {
    //create the hand controllers
    player1Controller = new HandController(InputDevice::ID_HAND_CONTROLLER_ONE);
    player2Controller = new HandController(InputDevice::ID_HAND_CONTROLLER_TWO);

    //init the memory bus
    memoryBus.init(0x10000);

    //8 bit RAM used by EXEC and cartridge
    RAM8bit.init(0x00F0, 0x0100, 8);
    memoryBus.addMemory(&RAM8bit);

    //the executive (EXEC) rom
    execROM.init(0x1000, 0x1000);
    memoryBus.addMemory(&execROM);

    //16 bit RAM used for...
    //    02F0       EXEC Object GRAM Allocation Table Pointer
    //    02F1-0318  1610 Stack
    //    0319-031C  16 bit RAM for EXEC
    //    031D-035C  Moving Object RAM Data Bases
    //    035D-035F  16 bit RAM for EXEC
    RAM16bit.init(0x0070, 0x02F0, 16);
    memoryBus.addMemory(&RAM16bit);

    //init the director
    emuDirector.init(this);
    addPeripheral(&emuDirector);

    //init the CPU
    cpu.init(&memoryBus, &SR1toINTRMLine, &SR2toBUSRQLine,
            &BUSAKtoSSTLine, 0x1000, 0x1004);
    processorBus.addProcessor(&cpu);

    //init the STIC
    stic.init(&cpu, &SR1toINTRMLine, &SR2toBUSRQLine, &BUSAKtoSSTLine);
    addPeripheral(&stic);

    //init the audio mixer
    am.init(&processorBus);
    processorBus.addProcessor(&am);

    //init the sound chip
    am.createLine(&psgAudioLine);
    psg.init(psgAudioLine, 0x01F0, player1Controller, player2Controller);
    psgAudioLine->acquire();
    addPeripheral(&psg);

    //init the ECS for ECS cartridge support
    am.createLine(&ecsAudioLine);
    this->ecs.init(ecsAudioLine);

    //init the Intellivoice for Intellivoice cartridge support
    am.createLine(&ivoiceAudioLine);
    this->intellivoice.init(ivoiceAudioLine);
}

/**
 * Turns the Intellivision on.
 */
INT32 Intellivision::turnOn() {
    if (on)
        return 0;

    if (!hasExec)
        return -1;

    if (!stic.hasGROMImage())
        return -1;

    if (currentCartridge->requiresECS() && !ecs.hasECSROMImage())
        return -1;

    reset();

    INT32 result;
    if (result = plugIn->init(this))
        return result;

    on = TRUE;
    return 0;
}

/**
 * Turns the Intellivision off.
 */
void Intellivision::turnOff() {
    if (!on)
        return;

    processorBus.stop();
    on = FALSE;
}

/**
 * Pushes the reset button on the Intellivision.
 */
void Intellivision::reset() {
    processorBus.reset();
    memoryBus.reset();
}

void Intellivision::run() {
    processorBus.run();
    plugIn->release(this);
}

/**
 * Inserts a cartridge into the Intellivision cartridge slot.
 *
 * @param cartridge the cartridge to insert
 */
void Intellivision::insertCartridge(Cartridge* cartridge) {
    //remove the old cartridge
    Cartridge* oldCart = currentCartridge;
    if (currentCartridge != NULL)
        removeCartridge();
    this->currentCartridge = cartridge;

    //add the ECS if this cartridge needs it
    if (currentCartridge->requiresECS()) {
        addPeripheral(&ecs);
        ecsAudioLine->acquire();
        ecsInUse = TRUE;
    }

    //add the Intellivoice if this cartridge can use it and we
    //have the ROM necessary for it
    if (currentCartridge->usesIntellivoice() &&
            intellivoice.hasIntellivoiceImage())
    {
        addPeripheral(&intellivoice);
        ivoiceAudioLine->acquire();
        intellivoiceInUse = TRUE;
    }

    //add the cartridge
    addPeripheral(cartridge);
}

Cartridge* Intellivision::getCartridge() {
    return currentCartridge;
}

/**
 * Removes the current cartridge from the Intellivision cartridge slot.
 */
void Intellivision::removeCartridge() {
    if (currentCartridge == NULL)
        return;

    Cartridge* oldCart = currentCartridge;
    removePeripheral(currentCartridge);

    //remove the Intellivoice
    if (intellivoiceInUse) {
        removePeripheral(&intellivoice);
        ivoiceAudioLine->release();
        intellivoiceInUse = FALSE;
    }

    if (ecsInUse) {
        removePeripheral(&ecs);
        ecsAudioLine->release();
        ecsInUse = TRUE;
    }

    currentCartridge = NULL;
}

void Intellivision::addPeripheral(Peripheral* p) {
    peripherals[periphCount] = p;
    periphCount++;

    INT32 i;
    INT32 count;
    count = p->getProcessorCount();
    for (i = 0; i < count; i++) {
        Processor* nextP;
        p->getProcessor(i, &nextP);
        processorBus.addProcessor(nextP);
    }

    count = p->getMemoryCount();
    for (i = 0; i < count; i++) {
        Memory* nextM;
        p->getMemory(i, &nextM);
        memoryBus.addMemory(nextM);
    }
}

void Intellivision::removePeripheral(Peripheral* p) {
    for (INT32 i = 0; i < periphCount; i++) {
        if (peripherals[i] == p) {
            for (INT32 j = i; j < (periphCount-1); j++)
                peripherals[j] = peripherals[j+1];
            periphCount--;

            INT32 count = p->getProcessorCount();
            for (INT32 i = 0; i < count; i++) {
                Processor* nextP;
                p->getProcessor(i, &nextP);
                processorBus.removeProcessor(nextP);
            }
        
            count = p->getMemoryCount();
            for (i = 0; i < count; i++) {
                Memory* nextM;
                p->getMemory(i, &nextM);
                memoryBus.removeMemory(nextM);
            }

            break;
        }
    }
}

void Intellivision::setExecImage(INT32* execImage) {
    execROM.loadImage(execImage);
    hasExec = TRUE;
}

BOOL Intellivision::hasExecImage() {
    return hasExec;
}

void Intellivision::setGROMImage(INT32* gromImage) {
    stic.setGROMImage(gromImage);
}

BOOL Intellivision::hasGROMImage() {
    return stic.hasGROMImage();
}

void Intellivision::setPlugIn(PlugIn* plugIn) {
    if (this->plugIn == plugIn)
        return;

    this->plugIn = plugIn;

    //give the plugIn over to the processor bus, since it's only
    //processors who need the plug-in
    processorBus.changePlugIn(plugIn);
}

PlugIn* Intellivision::getPlugIn() {
    return plugIn;
}

void Intellivision::enableIntellivoiceSupport(INT32* image) {
    intellivoice.setIntellivoiceImage(image);
}

void Intellivision::enableECSSupport(INT32* image) {
    ecs.setECSROMImage(image);
}
